@@ -52,6 +52,7 @@
 
 #include "shvar.h"
 #include "userhelper.h"
+#include "userhelper-messages.h"
 
 #ifdef DEBUG_USERHELPER
 #define debug_msg(...) g_print(__VA_ARGS__)
@@ -78,6 +79,7 @@ extern char **environ;
 
 /* A structure type which we use to carry psuedo-global data around with us. */
 struct app_data {
+	const struct pam_conv *conv;
 	pam_handle_t *pamh;
 	gboolean fallback_allowed, fallback_chosen, canceled;
 	FILE *input, *output;
@@ -88,6 +90,12 @@ struct app_data {
 	int sn_workspace;
 };
 
+/* A mixed-mode conversation function suitable for use with X.
+   data->conv->conv == converse_pipe is used to check whether we are run
+   by userhelper_runv() which e.g. writes error messages for us. */
+static int converse_pipe(int num_msg, const struct pam_message **msg,
+			 struct pam_response **resp, void *appdata_ptr);
+
 #ifdef WITH_SELINUX
 static int checkAccess(unsigned int selaccess) {
   int status=-1;
@@ -114,16 +122,33 @@ static int checkAccess(unsigned int selaccess) {
 }
 #endif /* WITH_SELINUX */
 
-/* Exit, returning the proper status code based on a PAM error code. */
+/* Exit, optionally writing a message to stderr. */
+static void G_GNUC_NORETURN
+die(struct app_data *data, int status)
+{
+	/* Be silent when status == 0, UNIX convention is to only report
+	   errors. */
+	if (status != 0 && data->conv->conv != converse_pipe) {
+		const char *msg;
+		enum uh_message_type type;
+
+		uh_exitstatus_message(status, &msg, &type);
+		if (type != UHM_SILENT)
+			fprintf(stderr, "%s\n", msg);
+	}
+	exit(status);
+}
+
+/* Exit, returning the proper status code based on a PAM error code.
+   Optionally write a message to stderr. */
 static int G_GNUC_NORETURN
 fail_exit(struct app_data *data, int pam_retval)
 {
 	int status;
 
 	/* This is a local error.  Bail. */
-	if (pam_retval == ERR_SHELL_INVALID) {
-		exit(ERR_SHELL_INVALID);
-	}
+	if (pam_retval == ERR_SHELL_INVALID)
+		die(data, ERR_SHELL_INVALID);
 
 	if (pam_retval == PAM_SUCCESS)
 		/* Just exit. */
@@ -172,7 +197,7 @@ fail_exit(struct app_data *data, int pam_retval)
 		}
 	}
 	debug_msg("userhelper: exiting with status %d.\n", status);
-	exit(status);
+	die(data, status);
 }
 
 /* Read a string from stdin, and return a freshly-allocated copy, without
@@ -760,12 +785,9 @@ err:
 }
 
 static int
-pipe_conv_exec_start(const struct pam_conv *conv)
+pipe_conv_exec_start(struct app_data *data)
 {
-	if (conv->conv == converse_pipe) {
-		struct app_data *data;
-
-		data = conv->appdata_ptr;
+	if (data->conv->conv == converse_pipe) {
 		/* There might have been no converse_pipe() call, so send the
 		   information now. */
 		send_sn_info(data);
@@ -786,12 +808,9 @@ pipe_conv_exec_start(const struct pam_conv *conv)
 }
 
 static void
-pipe_conv_exec_fail(const struct pam_conv *conv)
+pipe_conv_exec_fail(struct app_data *data)
 {
-	if (conv->conv == converse_pipe) {
-		struct app_data *data;
-
-		data = conv->appdata_ptr;
+	if (data->conv->conv == converse_pipe) {
 		debug_msg("userhelper: exec failed\n");
 		/* There might have been no converse_pipe() call, so send the
 		   information now, just to be sure. */
@@ -984,43 +1003,43 @@ is_grouplist_member(const char *username, const char * grouplist)
 }
 
 static void
-become_super_supplementary_groups(void)
+become_super_supplementary_groups(struct app_data *data)
 {
 	if (initgroups("root", 0) != 0) {
 		debug_msg("userhelper: initgroups() failure: %s\n",
 			  strerror(errno));
-		exit(ERR_EXEC_FAILED);
+		die(data, ERR_EXEC_FAILED);
 	}
 }
 
 static void
-become_super_other(void)
+become_super_other(struct app_data *data)
 {
 	/* Become the superuser.
 	   Yes, setuid() and friends can fail, even for superusers. */
 	if (setregid(0, 0) != 0 || setreuid(0, 0) != 0) {
 		debug_msg("userhelper: set*id() failure: %s\n",
 			  strerror(errno));
-		exit(ERR_EXEC_FAILED);
+		die(data, ERR_EXEC_FAILED);
 	}
 	if ((geteuid() != 0) ||
 	    (getuid() != 0) ||
 	    (getegid() != 0) ||
 	    (getgid() != 0)) {
 		debug_msg("userhelper: set*id() didn't work\n");
-		exit(ERR_EXEC_FAILED);
+		die(data, ERR_EXEC_FAILED);
 	}
 }
 
 static void
-become_super(void)
+become_super(struct app_data *data)
 {
-	become_super_supplementary_groups();
-	become_super_other();
+	become_super_supplementary_groups(data);
+	become_super_other(data);
 }
 
 static void
-become_normal(const char *user)
+become_normal(struct app_data *data, const char *user)
 {
 	gid_t gid;
 	uid_t uid;
@@ -1033,23 +1052,24 @@ become_normal(const char *user)
 	    setreuid(uid, uid) != 0) {
 		debug_msg("userhelper: set*id() failure: %s\n",
 			  strerror(errno));
-		exit(ERR_EXEC_FAILED);
+		die(data, ERR_EXEC_FAILED);
 	}
 	/* Verify that we're back to normal. */
 	if (getegid() != gid || getgid() != gid) {
 		debug_msg("userhelper: still setgid()\n");
-		exit(ERR_EXEC_FAILED);
+		die(data, ERR_EXEC_FAILED);
 	}
 	/* Yes, setuid() can fail. */
 	if (geteuid() != uid || getuid() != uid) {
 		debug_msg("userhelper: still setuid()\n");
-		exit(ERR_EXEC_FAILED);
+		die(data, ERR_EXEC_FAILED);
 	}
 }
 
-/* Determine the name of the user who ran userhelper. */
+/* Determine the name of the user who ran userhelper.
+   Exit on error, optionally writing to stderr. */
 static char *
-get_invoking_user(void)
+get_invoking_user(struct app_data *data)
 {
 	struct passwd *pwd;
 	char *ret=NULL;
@@ -1061,7 +1081,7 @@ get_invoking_user(void)
 	} else {
 		/* I have no name and I must have one. */
 		debug_msg("userhelper: i have no name\n");
-		exit(ERR_UNK_ERROR);
+		die(data, ERR_UNK_ERROR);
 	}
 
 	debug_msg("userhelper: ruid user = '%s'\n", ret);
@@ -1069,14 +1089,15 @@ get_invoking_user(void)
 	return ret;
 }
 
-/* Determine the name of the user as whom we must authenticate. */
+/* Determine the name of the user as whom we must authenticate.
+   Exit on error, optionally writing to stderr. */
 static char *
-get_user_for_auth(shvarFile *s)
+get_user_for_auth(struct app_data *data, shvarFile *s)
 {
 	char *ret;
 	char *invoking_user, *configured_user, *configured_asusergroups;
 
-	invoking_user = get_invoking_user();
+	invoking_user = get_invoking_user(data);
 
 	ret = NULL;
 
@@ -1100,9 +1121,9 @@ get_user_for_auth(shvarFile *s)
 				ret = invoking_user;
 			} else
 				ret = configured_user;
-		} else if (strcmp(configured_user, "<none>") == 0) {
-			exit(ERR_NO_RIGHTS);
-		} else
+		} else if (strcmp(configured_user, "<none>") == 0)
+			die(data, ERR_NO_RIGHTS);
+		else
 			ret = configured_user;
 		g_free(configured_asusergroups);
 	}
@@ -1118,7 +1139,8 @@ get_user_for_auth(shvarFile *s)
 	return NULL;
 }
 
-/* Set various attributes of DATA, including the requesting user USER. */
+/* Set various attributes of DATA, including the requesting user USER.
+   Exit on error, optionally writing to stderr. */
 static void
 set_pam_items(struct app_data *data, const char *user)
 {
@@ -1147,13 +1169,11 @@ set_pam_items(struct app_data *data, const char *user)
  * application data (which includes the ability to cancel if the user requests
  * it.  For this task, we don't retry on failure. */
 static void G_GNUC_NORETURN
-passwd(const char *user, struct pam_conv *conv)
+passwd(const char *user, struct app_data *data)
 {
 	int retval;
-	struct app_data *data;
 
-	data = conv->appdata_ptr;
-	retval = pam_start("passwd", user, conv, &data->pamh);
+	retval = pam_start("passwd", user, data->conv, &data->pamh);
 	if (retval != PAM_SUCCESS) {
 		debug_msg("userhelper: pam_start() failed\n");
 		fail_exit(data, retval);
@@ -1183,7 +1203,7 @@ passwd(const char *user, struct pam_conv *conv)
  * provide an interface to do this, because it's not PAM's job to manage this
  * stuff, so farm it out to a different library. */
 static void G_GNUC_NORETURN
-chfn(const char *user, struct pam_conv *conv, lu_prompt_fn *prompt,
+chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
      const char *new_full_name, const char *new_office,
      const char *new_office_phone, const char *new_home_phone,
      const char *new_shell)
@@ -1197,7 +1217,6 @@ chfn(const char *user, struct pam_conv *conv, lu_prompt_fn *prompt,
 	GValueArray *values;
 	GValue *value, val;
 	int tryagain = 3, retval;
-	struct app_data *data;
 	gboolean ret;
 
 	debug_msg("userhelper: chfn(\"%s\", \"%s\", \"%s\", \"%s\", "
@@ -1210,23 +1229,18 @@ chfn(const char *user, struct pam_conv *conv, lu_prompt_fn *prompt,
 
 	/* Verify that the fields we were given on the command-line
 	 * are sane (i.e., contain no forbidden characters). */
-	if (new_full_name && strpbrk(new_full_name, ":,=")) {
-		exit(ERR_FIELDS_INVALID);
-	}
-	if (new_office && strpbrk(new_office, ":,=")) {
-		exit(ERR_FIELDS_INVALID);
-	}
-	if (new_office_phone && strpbrk(new_office_phone, ":,=")) {
-		exit(ERR_FIELDS_INVALID);
-	}
-	if (new_home_phone && strpbrk(new_home_phone, ":,=")) {
-		exit(ERR_FIELDS_INVALID);
-	}
+	if (new_full_name && strpbrk(new_full_name, ":,="))
+		die(data, ERR_FIELDS_INVALID);
+	if (new_office && strpbrk(new_office, ":,="))
+		die(data, ERR_FIELDS_INVALID);
+	if (new_office_phone && strpbrk(new_office_phone, ":,="))
+		die(data, ERR_FIELDS_INVALID);
+	if (new_home_phone && strpbrk(new_home_phone, ":,="))
+		die(data, ERR_FIELDS_INVALID);
 
 	/* Start up PAM to authenticate the user, this time pretending
 	 * we're "chfn". */
-	data = conv->appdata_ptr;
-	retval = pam_start("chfn", user, conv, &data->pamh);
+	retval = pam_start("chfn", user, data->conv, &data->pamh);
 	if (retval != PAM_SUCCESS) {
 		debug_msg("userhelper: pam_start() failed\n");
 		fail_exit(data, retval);
@@ -1265,7 +1279,7 @@ chfn(const char *user, struct pam_conv *conv, lu_prompt_fn *prompt,
 	if (strcmp(user, authed_user) != 0) {
 		debug_msg("userhelper: username(%s) != authuser(%s)\n",
 			  user, authed_user);
-		exit(ERR_UNK_ERROR);
+		die(data, ERR_UNK_ERROR);
 	}
 
 	/* Check if the user is allowed to change her information at
@@ -1355,7 +1369,7 @@ chfn(const char *user, struct pam_conv *conv, lu_prompt_fn *prompt,
 		lu_ent_free(ent);
 		lu_end(context);
 		pam_end(data->pamh, PAM_ABORT);
-		exit(ERR_FIELDS_INVALID);
+		die(data, ERR_FIELDS_INVALID);
 	}
 
 	/* Build a new value for the GECOS data. */
@@ -1462,7 +1476,7 @@ construct_cmdline(const char *argv0, char **argv)
 
 static void G_GNUC_NORETURN
 wrap(const char *user, const char *program,
-     struct pam_conv *conv, struct pam_conv *text_conv, lu_prompt_fn *prompt,
+     struct app_data *data, struct pam_conv *text_conv, lu_prompt_fn *prompt,
      int argc, char **argv)
 {
 	/* We're here to wrap the named program.  After authenticating as the
@@ -1481,7 +1495,6 @@ wrap(const char *user, const char *program,
 	int session, tryagain, gui, retval;
 	struct stat sbuf;
 	struct passwd *pwd;
-	struct app_data *data;
 	shvarFile *s;
 
 	(void)prompt;
@@ -1505,7 +1518,7 @@ wrap(const char *user, const char *program,
 	    (sbuf.st_mode & S_IWOTH)) {
 		debug_msg("userhelper: bad file permissions: %s \n",
 			  apps_filename);
-		exit(ERR_UNK_ERROR);
+		die(data, ERR_UNK_ERROR);
 	}
 	g_free(apps_filename);
 
@@ -1633,34 +1646,6 @@ wrap(const char *user, const char *program,
 	/* Pass the original UID to the new program */
 	setenv("USERHELPER_UID", g_strdup_printf("%jd", (intmax_t)getuid()), 1);
 
-	data = conv->appdata_ptr;
-	user_pam = get_user_for_auth(s);
-
-	/* Read the path to the program to run. */
-	constructed_path = svGetValue(s, "PROGRAM");
-	if (!constructed_path || constructed_path[0] != '/') {
-		g_free(constructed_path);
-		/* Criminy....  The system administrator didn't give us an
-		 * absolute path to the program!  Guess either /usr/sbin or
-		 * /sbin, and then give up if we don't find anything by that
-		 * name in either of those directories.  FIXME: we're a setuid
-		 * app, so access() may not be correct here, as it may give
-		 * false negatives.  But then, it wasn't an absolute path. */
-		constructed_path = g_strconcat("/usr/sbin/", program, NULL);
-		if (access(constructed_path, X_OK) != 0) {
-			/* Try the second directory. */
-			strcpy(constructed_path, "/sbin/");
-			strcat(constructed_path, program);
-			if (access(constructed_path, X_OK)) {
-				/* Nope, not there, either. */
-				debug_msg("userhelper: couldn't find wrapped "
-					  "binary\n");
-				exit(ERR_NO_PROGRAM);
-			}
-		}
-	}
-
-
 	/* We can forcefully disable the GUI from the configuration
 	 * file (a la blah-nox applications). */
 	gui = svTrueValue(s, "GUI", TRUE);
@@ -1685,18 +1670,43 @@ wrap(const char *user, const char *program,
 		/* We are not really executing anything yet, but this switches
 		   off the parent to a "pass exit code through" mode without
 		   displaying any unwanted GUI dialogs. */
-		retval = pipe_conv_exec_start(conv);
+		retval = pipe_conv_exec_start(data);
 		if (retval != 0)
-			exit(retval);
-		conv = text_conv;
+			die(data, retval);
+		data->conv = text_conv;
 	}
 
+	/* Read the path to the program to run. */
+	constructed_path = svGetValue(s, "PROGRAM");
+	if (!constructed_path || constructed_path[0] != '/') {
+		g_free(constructed_path);
+		/* Criminy....  The system administrator didn't give us an
+		 * absolute path to the program!  Guess either /usr/sbin or
+		 * /sbin, and then give up if we don't find anything by that
+		 * name in either of those directories.  FIXME: we're a setuid
+		 * app, so access() may not be correct here, as it may give
+		 * false negatives.  But then, it wasn't an absolute path. */
+		constructed_path = g_strconcat("/usr/sbin/", program, NULL);
+		if (access(constructed_path, X_OK) != 0) {
+			/* Try the second directory. */
+			strcpy(constructed_path, "/sbin/");
+			strcat(constructed_path, program);
+			if (access(constructed_path, X_OK)) {
+				/* Nope, not there, either. */
+				debug_msg("userhelper: couldn't find wrapped "
+					  "binary\n");
+				die(data, ERR_NO_PROGRAM);
+			}
+		}
+	}
+
+	user_pam = get_user_for_auth(data, s);
 	/* Verify that the user we need to authenticate as has a home
 	 * directory. */
 	pwd = getpwnam(user_pam);
 	if (pwd == NULL) {
 		debug_msg("userhelper: no user named %s exists\n", user_pam);
-		exit(ERR_NO_USER);
+		die(data, ERR_NO_USER);
 	}
 
 	/* If the user we're authenticating as has root's UID, then it's
@@ -1765,7 +1775,7 @@ wrap(const char *user, const char *program,
 	svCloseFile(s);
 
 	/* Start up PAM to authenticate the specified user. */
-	retval = pam_start(program, user_pam, conv, &data->pamh);
+	retval = pam_start(program, user_pam, data->conv, &data->pamh);
 	if (retval != PAM_SUCCESS) {
 		debug_msg("userhelper: pam_start() failed\n");
 		fail_exit(data, retval);
@@ -1793,7 +1803,7 @@ wrap(const char *user, const char *program,
 			 * application can run normally. */
 			argv[optind - 1] = strdup(program);
 			environ = environ_save;
-			become_normal(user);
+			become_normal(data, user);
 			if (data->input != NULL) {
 				fflush(data->input);
 				fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC);
@@ -1802,9 +1812,9 @@ wrap(const char *user, const char *program,
 				fflush(data->output);
 				fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC);
 			}
-			retval = pipe_conv_exec_start(conv);
+			retval = pipe_conv_exec_start(data);
 			if (retval != 0)
-				exit(retval);
+				die(data, retval);
 #ifdef USE_STARTUP_NOTIFICATION
 			if (data->sn_id) {
 				debug_msg("userhelper: setting "
@@ -1815,8 +1825,8 @@ wrap(const char *user, const char *program,
 			}
 #endif
 			execv(constructed_path, argv + optind - 1);
-			pipe_conv_exec_fail(conv);
-			exit(ERR_EXEC_FAILED);
+			pipe_conv_exec_fail(data);
+			die(data, ERR_EXEC_FAILED);
 		} else {
 			/* Well, we tried. */
 			fail_exit(data, retval);
@@ -1830,9 +1840,8 @@ wrap(const char *user, const char *program,
 		pam_end(data->pamh, retval);
 		fail_exit(data, retval);
 	}
-	if (strcmp(user_pam, auth_user) != 0) {
-		exit(ERR_UNK_ERROR);
-	}
+	if (strcmp(user_pam, auth_user) != 0)
+		die(data, ERR_UNK_ERROR);
 
 	/* Verify that the authenticated user is allowed to run this
 	 * service now. */
@@ -1847,7 +1856,7 @@ wrap(const char *user, const char *program,
 	pwd = getpwnam(user_pam);
 	if (pwd == NULL) {
 		debug_msg("userhelper: no user named %s exists\n", user_pam);
-		exit(ERR_NO_USER);
+		die(data, ERR_NO_USER);
 	}
 
 	/* What we do now depends on whether or not we need to open
@@ -1869,7 +1878,7 @@ wrap(const char *user, const char *program,
 			fail_exit(data, retval);
 		}
 
-		become_super_supplementary_groups();
+		become_super_supplementary_groups(data);
 
 		retval = pam_setcred(data->pamh, PAM_ESTABLISH_CRED);
 		if (retval != PAM_SUCCESS) {
@@ -1879,9 +1888,8 @@ wrap(const char *user, const char *program,
 
 		/* Start up a child process we can wait on. */
 		child = fork();
-		if (child == -1) {
-			exit(ERR_EXEC_FAILED);
-		}
+		if (child == -1)
+			die(data, ERR_EXEC_FAILED);
 		if (child == 0) {
 			/* We're in the child.  Make a few last-minute
 			 * preparations and exec the program. */
@@ -1900,7 +1908,7 @@ wrap(const char *user, const char *program,
 				  constructed_path);
 			/* become_super_supplementary_groups() called in the
 			   parent. */
-			become_super_other();
+			become_super_other(data);
 			if (data->input != NULL) {
 				fflush(data->input);
 				fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC);
@@ -1909,9 +1917,9 @@ wrap(const char *user, const char *program,
 				fflush(data->output);
 				fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC);
 			}
-			retval = pipe_conv_exec_start(conv);
+			retval = pipe_conv_exec_start(data);
 			if (retval != 0)
-				exit(retval);
+				die(data, retval);
 #ifdef USE_STARTUP_NOTIFICATION
 			if (data->sn_id) {
 				debug_msg("userhelper: setting "
@@ -1933,8 +1941,8 @@ wrap(const char *user, const char *program,
 			syslog(LOG_ERR, "could not run '%s' with "
 			       "root privileges on behalf of '%s': %s",
 			       cmdline, user, strerror(errno));
-			pipe_conv_exec_fail(conv);
-			exit(ERR_EXEC_FAILED);
+			pipe_conv_exec_fail(data);
+			die(data, ERR_EXEC_FAILED);
 		}
 		/* We're in the parent.  Wait for the child to exit.  The child
 		   is calling pipe_conv_exec_{start,fail} () to define the
@@ -1948,7 +1956,7 @@ wrap(const char *user, const char *program,
 		/* Close the session. */
 		retval = pam_close_session(data->pamh, 0);
 		if (retval != PAM_SUCCESS) {
-			pipe_conv_exec_fail(conv);
+			pipe_conv_exec_fail(data);
 			pam_setcred(data->pamh, PAM_DELETE_CRED);
 			pam_end(data->pamh, retval);
 			fail_exit(data, retval);
@@ -1971,7 +1979,7 @@ wrap(const char *user, const char *program,
 		argv[optind - 1] = strdup(program);
 		debug_msg("userhelper: about to exec \"%s\"\n",
 			  constructed_path);
-		become_super();
+		become_super(data);
 		if (data->input != NULL) {
 			fflush(data->input);
 			fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC);
@@ -1980,9 +1988,9 @@ wrap(const char *user, const char *program,
 			fflush(data->output);
 			fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC);
 		}
-		retval = pipe_conv_exec_start(conv);
+		retval = pipe_conv_exec_start(data);
 		if (retval != 0)
-			exit(retval);
+			die(data, retval);
 #ifdef USE_STARTUP_NOTIFICATION
 		if (data->sn_id) {
 			debug_msg("userhelper: setting "
@@ -2002,8 +2010,8 @@ wrap(const char *user, const char *program,
 		syslog(LOG_ERR, "could not run '%s' with "
 		       "root privileges on behalf of '%s': %s",
 		       cmdline, user, strerror(errno));
-		pipe_conv_exec_fail(conv);
-		exit(ERR_EXEC_FAILED);
+		pipe_conv_exec_fail(data);
+		die(data, ERR_EXEC_FAILED);
 	}
 }
 
@@ -2037,6 +2045,7 @@ main(int argc, char **argv)
 
 	/* State variable we pass around. */
 	struct app_data app_data = {
+		NULL,
 		NULL,
 		FALSE, FALSE, FALSE,
 		NULL, NULL,
@@ -2060,7 +2069,6 @@ main(int argc, char **argv)
 		converse_console,
 		&app_data,
 	};
-	struct pam_conv *conv;
 
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
@@ -2153,11 +2161,11 @@ main(int argc, char **argv)
 		if (isatty(STDIN_FILENO)) {
 			/* We have a controlling tty on which we can disable
 			 * echoing, so use the text conversation method. */
-			conv = &text_conv;
+			app_data.conv = &text_conv;
 		} else {
 			/* We have no controlling terminal -- being run from
 			 * cron or some other mechanism? */
-			conv = &silent_conv;
+			app_data.conv = &silent_conv;
 #if 0
 			/* FIXME: print a warning here? */
 			fprintf(stderr, _("Unable to open graphical window, "
@@ -2174,11 +2182,11 @@ main(int argc, char **argv)
 			debug_msg("userhelper: invalid call\n");
 			exit(ERR_INVALID_CALL);
 		}
-		conv = &pipe_conv;
+		app_data.conv = &pipe_conv;
 		prompt = &prompt_pipe;
 	}
 
-	user_name = get_invoking_user();
+	user_name = get_invoking_user(&app_data);
 	debug_msg("userhelper: current user is %s\n", user_name);
 
 	/* If we didn't get the -w flag, the last argument can be a user's
@@ -2206,7 +2214,7 @@ main(int argc, char **argv)
 				       "SELinux context %s is not allowed to change information for user \"%s\"\n",
 				       context, user_name);
 				g_free(user_name);
-				exit(ERR_NO_USER);
+				die(&app_data, ERR_NO_USER);
 			}
 #endif
 			debug_msg("userhelper: modifying account data for %s\n",
@@ -2218,19 +2226,19 @@ main(int argc, char **argv)
 		if ((pwd == NULL) || (pwd->pw_name == NULL)) {
 			debug_msg("userhelper: user %s doesn't exist\n",
 				  user_name);
-			exit(ERR_NO_USER);
+			die(&app_data, ERR_NO_USER);
 		}
 	}
 
 	/* Change password? */
 	if (c_flag) {
-		passwd(user_name, conv);
+		passwd(user_name, &app_data);
 		g_assert_not_reached();
 	}
 
 	/* Change GECOS data or shell? */
 	if (SHELL_FLAGS) {
-		chfn(user_name, conv, prompt,
+		chfn(user_name, &app_data, prompt,
 		     new_full_name, new_office,
 		     new_office_phone, new_home_phone,
 		     new_shell);
@@ -2239,7 +2247,7 @@ main(int argc, char **argv)
 
 	/* Wrap some other program? */
 	if (w_flag) {
-		wrap(user_name, wrapped_program, conv, &text_conv, prompt,
+		wrap(user_name, wrapped_program, &app_data, &text_conv, prompt,
 		     argc, argv);
 		g_assert_not_reached();
 	}
